#
# This file is part of Apollo.
#
# Copyright (c) 2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Controllers for communicating with Apollo through the FPGA_ADV pin """

from amaranth                       import Elaboratable, Module, Signal, Mux

from luna.gateware.usb.usb2.request import USBRequestHandler
from usb_protocol.types             import USBRequestType, USBRequestRecipient


class ApolloAdvertiser(Elaboratable):
    """ Gateware that implements an announcement signal for Apollo using the FPGA_ADV pin.

    Used to tell Apollo that the gateware wants to use the CONTROL port.
    Apollo will keep the port switch connected to the FPGA after a reset as long as this 
    signal is being received and the port takeover is allowed.

    I/O ports:
        I: stop -- Advertisement signal is stopped if this line is asserted.
    """
    def __init__(self, pad=None, clk_freq_hz=None):
        self.pad         = pad
        self.clk_freq_hz = clk_freq_hz
        self.stop        = Signal()

    def default_request_handler(self, if_number):
        return ApolloAdvertiserRequestHandler(if_number, self.stop)

    def elaborate(self, platform):
        m = Module()

        # Handle default values.
        if self.pad is None:
            self.pad = platform.request("int")
        if self.clk_freq_hz is None:
            self.clk_freq_hz = platform.DEFAULT_CLOCK_FREQUENCIES_MHZ["sync"] * 1e6

        # Generate clock with 20ms period.
        half_period = int(self.clk_freq_hz * 10e-3)
        timer       = Signal(range(half_period))
        clk         = Signal()
        m.d.sync   += timer.eq(Mux(timer == half_period-1, 0, timer+1))
        with m.If((timer == 0) & (~self.stop)):
            m.d.sync += clk.eq(~clk)

        # Drive the FPGA_ADV pin with the generated clock signal.
        m.d.comb += self.pad.o.eq(clk)
        
        return m


class ApolloAdvertiserRequestHandler(USBRequestHandler):
    """ Request handler for ApolloAdvertiser. 
    
    Implements default vendor requests related to ApolloAdvertiser.
    """

    """ The bInterfaceProtocol version supported by this request handler. """
    PROTOCOL_VERSION = 0x00

    REQUEST_APOLLO_ADV_STOP = 0xF0

    def __init__(self, if_number, stop_pin):
        super().__init__()
        self.if_number = if_number
        self.stop_pin  = stop_pin

    def elaborate(self, platform):
        m = Module()

        interface         = self.interface
        setup             = self.interface.setup

        #
        # Vendor request handlers.

        with m.If((setup.type == USBRequestType.VENDOR) & \
                  (setup.recipient == USBRequestRecipient.INTERFACE) & \
                  (setup.index == self.if_number)):
            
            with m.If(setup.request == self.REQUEST_APOLLO_ADV_STOP):

                # Notify that we want to manage this request
                m.d.comb += interface.claim.eq(1)

                # Once the receive is complete, respond with an ACK.
                with m.If(interface.rx_ready_for_response):
                    m.d.comb += interface.handshakes_out.ack.eq(1)

                # If we reach the status stage, send a ZLP.
                with m.If(interface.status_requested):
                    m.d.comb += self.send_zlp()
                    m.d.usb += self.stop_pin.eq(1)

        return m



#
# This file is part of Apollo.
#
# Copyright (c) 2020-2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

import os
import time
import usb.core
import platform
import errno
import sys

from .jtag  import JTAGChain
from .spi   import DebugSPIConnection
from .ila   import ApolloILAFrontend
from .ecp5  import ECP5_JTAGProgrammer, ECP5_JTAGDebugSPIConnection, ECP5_JTAGRegisters
from .intel import IntelJTAGProgrammer

from .onboard_jtag import *

import importlib.metadata
__version__ = importlib.metadata.version(__package__)


class DebuggerNotFound(IOError):
    pass


def create_ila_frontend(ila, *, use_cs_multiplexing=False):
    """ Convenience method that instantiates an Apollo debug session and creates an ILA frontend from it.

    Parameters:
        ila -- The SyncSerialILA object we'll be connecting to.
    """
    debugger = ApolloDebugger()
    return ApolloILAFrontend(debugger, ila=ila, use_inverted_cs=use_cs_multiplexing)



class ApolloDebugger:
    """ Class representing a link to an Apollo Debug Module. """

    # VID/PID pairs for Apollo and gateware.
    APOLLO_USB_IDS = [(0x1d50, 0x615c), (0x1209, 0x0010)]
    LUNA_USB_IDS   = [(0x1d50, 0x615b)]

    # Add pid.codes test VID/PID pairs with PID from 0x0001 to 0x0005 used in
    # LUNA example gateware.
    for i in range(5):
        LUNA_USB_IDS += [(0x1209, i+1)]

    # Add pid.codes test VID/PID used by flash bridge.
    LUNA_USB_IDS += [(0x1209, 0x000f)]

    # If we have a LUNA_USB_IDS variable, we can use it to find the LUNA device.
    if os.getenv("LUNA_USB_IDS"):
        LUNA_USB_IDS += [tuple([int(x, 16) for x in os.getenv("LUNA_USB_IDS").split(":")])]

    REQUEST_GET_ID                  = 0xa0
    REQUEST_SET_LED_PATTERN         = 0xa1
    REQUEST_GET_FIRMWARE_VERSION    = 0xa2
    REQUEST_GET_USB_API_VERSION     = 0xa3
    REQUEST_RECONFIGURE             = 0xc0
    REQUEST_FORCE_FPGA_OFFLINE      = 0xc1
    REQUEST_ALLOW_FPGA_TAKEOVER_USB = 0xc2

    LED_PATTERN_IDLE = 500
    LED_PATTERN_UPLOAD = 50


    # External boards (non-Cynthion boards) are indicated with a Major revision of 0xFF.
    # Their minor revision then encodes the board type.
    EXTERNAL_BOARD_MAJOR = 0xFF
    EXTERNAL_BOARD_NAMES = {
        0: "Daisho [rev 31-Oct-2014]",
        1: "Xil.se Pergola FPGA",
        2: "Adafruit QT Py ATSAMD21E18",
    }

    EXTERNAL_BOARD_PROGRAMMERS = {
        0: IntelJTAGProgrammer,
        2: ECP5_JTAGProgrammer,
    }


    # Cynthion subdevices (specialized variants Cynthion) use a major of 0xFE.
    SUBDEVICE_MAJORS = {
        0xFE: "Amalthea"
    }

    backend = None


    def __init__(self, force_offline=False, device=None):
        """ Sets up a connection to the debugger. """

        # Try to create a connection to our Apollo debug firmware.
        if device is None:
            device = self._find_device(self.APOLLO_USB_IDS, custom_match=self._device_has_apollo_id)

        # If Apollo VID/PID is not found, try to find a gateware VID/PID with a valid Apollo stub
        # interface. If found, request the gateware to liberate the USB port. In devices with a
        # shared port, this effectively hands off the USB port to Apollo.
        if device is None:

            # First, find the candidate device...
            fpga_device = self._find_device(self.LUNA_USB_IDS, custom_match=self._device_has_stub_iface)
            if fpga_device is None:
                raise DebuggerNotFound("No Apollo debugger or stub interface found.")
            elif not force_offline:
                raise DebuggerNotFound("Apollo stub interface found but not requested to be forced offline.")

            # ... and now request a USB handoff to Apollo
            try:
                self._request_handoff(fpga_device)
            except usb.USBError as e:
                raise DebuggerNotFound(f"Handoff request failed: {e.strerror}")

            # Wait for Apollo to enumerate and try again
            device = self._find_device(self.APOLLO_USB_IDS, custom_match=self._device_has_apollo_id, timeout=5000)
            if device is None:
                raise DebuggerNotFound("Handoff was requested, but Apollo debugger not found.")

        self.device = device
        self.major, self.minor = self.get_hardware_revision()

        # Create a basic JTAG chain interface, for debugging convenience.
        self.jtag  = JTAGChain(self)

        # Try to create an SPI-over-JTAG tunnel, if this board supports it.
        self.spi, self.registers = self.create_jtag_spi(self.jtag)

        # If it doesn't, use a hard SPI for JTAG-SPI.
        if self.spi is None:
            self.spi   = DebugSPIConnection(self)
            self.registers = self.spi

    @classmethod
    def _request_handoff(cls, device):
        """ Requests the gateware to liberate the USB port. """
        # Find the Apollo stub interface first
        stub_if = cls._device_has_stub_iface(device, return_iface=True)
        if stub_if is None:
            raise DebuggerNotFound("No Apollo stub interface found")

        # In Windows, we first need to claim the target interface.
        if platform.system() == "Windows":
            usb.util.claim_interface(device, stub_if)

        # Send the request
        intf_number = stub_if.bInterfaceNumber
        REQUEST_APOLLO_ADV_STOP = 0xF0
        request_type = usb.ENDPOINT_OUT | usb.RECIP_INTERFACE | usb.TYPE_VENDOR
        device.ctrl_transfer(request_type, REQUEST_APOLLO_ADV_STOP, wIndex=intf_number, timeout=5000)

    @classmethod
    def _init_backend(cls):
        """Initialize the USB backend."""
        if cls.backend is None:
            import usb.backend.libusb1

            # In Windows, we need to specify the libusb library location to create a backend.
            if platform.system() == "Windows":
                # Determine the path to libusb-1.0.dll.
                try:
                    from importlib_resources import files # <= 3.8
                except:
                    from importlib.resources import files # >= 3.9
                libusb_dll = os.path.join(files("usb1"), "libusb-1.0.dll")

                # Create a backend by explicitly passing the path to libusb_dll.
                cls.backend = usb.backend.libusb1.get_backend(find_library=lambda x: libusb_dll)
            else:
                # On other systems we can just use the default backend.
                cls.backend = usb.backend.libusb1.get_backend()

    @classmethod
    def _find_device(cls, ids, custom_match=None, timeout=0):
        """Find a USB device matching a list of VID/PIDs."""
        cls._init_backend()
        wait = 0
        while True:
            for vid, pid in ids:
                device = usb.core.find(backend=cls.backend, idVendor=vid, idProduct=pid, custom_match=custom_match)
                if device is not None:
                    return device
            # Should we wait and try again?
            if wait >= timeout:
                break
            time.sleep(0.5)
            wait += 500
        return None

    @classmethod
    def _find_all_devices(cls, ids, custom_match=None):
        """Find all USB devices matching a list of VID/PIDs."""
        cls._init_backend()
        devices = []
        candidates = []

        for vid, pid in ids:
            candidates.extend(usb.core.find(True, cls.backend, idVendor=vid, idProduct=pid))

        for device in candidates:
            try:
                if custom_match is None or custom_match(device):
                    devices.append(device)
            except (usb.USBError, NotImplementedError):
                # A permissions error or NotImplementedError is likely on
                # Windows if the device is not the expected type of device.
                # Other typical errors are transient conditions shortly after
                # device enumeration of a device that is not yet ready.
                pass
        return devices

    @staticmethod
    def _device_has_stub_iface(device, return_iface=False):
        """Check if a device has an Apollo stub interface present.

        Optionally return the interface itself.
        """
        for cfg in device:
            stub_if = usb.util.find_descriptor(cfg, bInterfaceClass=0xFF, bInterfaceSubClass=0x00)
            if stub_if is not None:
                return stub_if if return_iface else True
        return None if return_iface else False

    @staticmethod
    def _device_has_apollo_id(device):
        """Check if a device identifies itself as an Apollo debugger."""
        request_type = usb.ENDPOINT_IN | usb.RECIP_DEVICE | usb.TYPE_VENDOR
        try:
            response = device.ctrl_transfer(request_type, ApolloDebugger.REQUEST_GET_ID, data_or_wLength=256, timeout=500)
        except usb.USBError as e:
            if e.errno == errno.EPIPE:
                # A pipe error occurs when the device does not implement a
                # vendor request with a number matching REQUEST_GET_ID. This is
                # the expected result if the device is running Saturn-V.
                return False
            else:
                raise
        finally:
            usb.util.dispose_resources(device)
        apollo_id = bytes(response).decode('utf-8').split('\x00')[0]
        return True if "Apollo" in apollo_id else False

    def detect_connected_version(self):
        """ Attempts to determine the revision of the connected hardware.

        Returns the relevant hardware's revision number, as (major, minor).
        """

        # Extract the major and minor from the device's USB descriptor.
        minor = self.device.bcdDevice & 0xFF
        major = self.device.bcdDevice >> 8

        return major, minor


    def get_fpga_type(self):
        """ Returns a string indicating the type of FPGA populated on the connected Apollo board.

        The returned format is the same as used in a nMigen platform file; and can be used to override
        a platform's device type.
        """

        with self.jtag as jtag:

            # First, we'll detect all devices on our JTAG chain.
            jtag_devices = jtag.enumerate()
            if not jtag_devices:
                raise IOError("Could not detect an FPGA via JTAG!")

            # ... and grab its device identifier.
            first_device = jtag_devices[0]
            if not hasattr(first_device, 'DEVICE'):
                raise IOError("First JTAG device in chain does not provide an FPGA type. Is this a proper board?")

            return first_device.DEVICE


    @property
    def serial_number(self):
        """ Returns the device's serial number, as a string. """
        return self.device.serial_number


    def get_hardware_revision(self):
        """ Returns the (major, minor) of the attached hardware revision. """

        minor = self.device.bcdDevice & 0xFF
        major = self.device.bcdDevice >> 8
        return major, minor


    def get_hardware_name(self):
        """ Returns a string describing this piece of hardware. """

        # If this is a non-Cynthion board, we'll look up its name in our table.
        if self.major == self.EXTERNAL_BOARD_MAJOR:
            return self.EXTERNAL_BOARD_NAMES[self.minor]

        # If this is a non-Cynthion board, we'll look up its name in our table.
        if self.major in self.SUBDEVICE_MAJORS:
            product_name = self.SUBDEVICE_MAJORS[self.major]
            major        = 0 # For now?
        else:
            product_name = "Cynthion"
            major        = self.major

        # Otherwise, identify it by its revision number.
        return f"{product_name} r{major}.{self.minor}"


    def get_compatibility_string(self):
        """ Returns 'Cynthion' for a Cynthion board; or 'Apollo' for supported external board."""

        if self.major == self.EXTERNAL_BOARD_MAJOR:
            return 'Apollo'
        elif self.major in self.SUBDEVICE_MAJORS:
            return self.SUBDEVICE_MAJORS[self.major]

        return 'Cynthion'


    def create_jtag_programmer(self, jtag_chain):
        """ Returns the JTAG programmer for the given device. """

        # If this is an external programmer, return its programmer type.
        if self.major == self.EXTERNAL_BOARD_MAJOR:
            programmer = self.EXTERNAL_BOARD_PROGRAMMERS[self.minor]

        # Otherwise, it should be an ECP5.
        else:
            programmer = ECP5_JTAGProgrammer

        return programmer(jtag_chain)



    def create_jtag_spi(self, jtag_chain):
        """ Returns a JTAG-over-SPI connection for the given device. """

        # If this is an external programmer, we don't yet know how to create a JTAG-SPI
        # interface for it. For now, assume we can't.
        if self.major == self.EXTERNAL_BOARD_MAJOR:
            return None, None

        # Use a real debug SPI on r0.1 and r0.2.
        elif self.major == 0 and self.minor < 3:
            return None, None

        # Otherwise, if we have a revision greater than r0.2, our SPI should be via JTAG.
        else:
            return ECP5_JTAGDebugSPIConnection(jtag_chain, self), ECP5_JTAGRegisters(jtag_chain)



    def out_request(self, number, value=0, index=0, data=None, timeout=500):
        """ Helper that issues an OUT control request to the debugger. """

        request_type = usb.ENDPOINT_OUT | usb.RECIP_DEVICE | usb.TYPE_VENDOR
        return self.device.ctrl_transfer(request_type, number, value, index, data, timeout=timeout)


    def in_request(self, number, value=0, index=0, length=0, timeout=500):
        """ Helper that issues an IN control request to the debugger. """

        request_type = usb.ENDPOINT_IN | usb.RECIP_DEVICE | usb.TYPE_VENDOR
        result = self.device.ctrl_transfer(request_type, number, value, index, length, timeout=timeout)

        return bytes(result)


    def set_led_pattern(self, number):
        self.out_request(self.REQUEST_SET_LED_PATTERN, number)


    def soft_reset(self):
        """ Resets the target (FPGA/etc) connected to the debug controller. """
        self.out_request(self.REQUEST_RECONFIGURE)


    def force_fpga_offline(self):
        """ Resets the target (FPGA/etc) connected to the debug controller. """
        self.out_request(self.REQUEST_FORCE_FPGA_OFFLINE)

    def allow_fpga_takeover_usb(self):
        """ Request Apollo to allow FPGA takeover of the USB port. Useful after reconfiguration. """
        self.out_request(self.REQUEST_ALLOW_FPGA_TAKEOVER_USB)

    def close(self):
        """ Closes the USB device so it can be reused, possibly by another ApolloDebugger """

        usb.util.dispose_resources(self.device)


    def get_firmware_version(self):
        c_string = self.in_request(self.REQUEST_GET_FIRMWARE_VERSION, length=256)
        return c_string.decode('utf-8').split('\x00')[0]


    def get_usb_api_version(self):
        raw_api_version = self.in_request(self.REQUEST_GET_USB_API_VERSION, length=2)
        api_major = int(raw_api_version[0])
        api_minor = int(raw_api_version[1])
        return (api_major, api_minor)


    def get_usb_api_version_string(self):
        (api_major, api_minor) = self.get_usb_api_version()
        return (f"{api_major}.{api_minor}")

    @classmethod
    def print_info(cls, ids=None, stub_ids=None, force_offline=False, timeout=5000, out=print):
        """ Print information about Apollo and all connected Apollo devices.

        Return True if any connected device is found.
        """
        out(f"Apollo version: {__version__}")
        out(f"Python version: {sys.version}\n")

        found_device = False
        if ids is None:
            ids = cls.APOLLO_USB_IDS
        if stub_ids is None:
            stub_ids = cls.LUNA_USB_IDS

        # Look for devices with stub interfaces.
        stub_devs = cls._find_all_devices(stub_ids, cls._device_has_stub_iface)

        for device in stub_devs:
            found_device = True
            out("Found Apollo stub interface!")
            out(f"\tBitstream: {device.product} ({device.manufacturer})")
            out(f"\tVendor ID: {device.idVendor:04x}")
            out(f"\tProduct ID: {device.idProduct:04x}")
            out(f"\tbcdDevice: {device.bcdDevice:04x}")
            out(f"\tBitstream serial number: {device.serial_number}")
            out("")

        count = 0
        if force_offline:
            apollo_devs = cls._find_all_devices(ids, cls._device_has_apollo_id)
            count = len(apollo_devs) + len(stub_devs)
            for device in apollo_devs:
                usb.util.dispose_resources(device)
            if count > 0:
                out(f"Forcing offline.\n")
            for device in stub_devs:
                try:
                    cls._request_handoff(device)
                except usb.USBError as e:
                    raise DebuggerNotFound(f"Handoff request failed: {e.strerror}")

        # Look for Apollo debuggers.
        start = time.time()
        while True:
            apollo_devs = cls._find_all_devices(ids, cls._device_has_apollo_id)
            if len(apollo_devs) >= count:
                break;
            if ((time.time() - start) * 1000) >= timeout:
                raise DebuggerNotFound("Handoff was requested, but Apollo debugger not found.")
            time.sleep(0.1)

        for device in apollo_devs:
            found_device = True
            debugger = ApolloDebugger(device=device)
            out(f"Found {debugger.get_compatibility_string()} device!")
            out(f"\tHardware: {debugger.get_hardware_name()}")
            out(f"\tManufacturer: {device.manufacturer}")
            out(f"\tProduct: {device.product}")
            out(f"\tSerial number: {device.serial_number}")
            out(f"\tVendor ID: {device.idVendor:04x}")
            out(f"\tProduct ID: {device.idProduct:04x}")
            out(f"\tbcdDevice: {device.bcdDevice:04x}")
            out(f"\tFirmware version: {debugger.get_firmware_version()}")
            out(f"\tUSB API version: {debugger.get_usb_api_version_string()}")
            if force_offline:
                debugger.force_fpga_offline()
                with debugger.jtag as jtag:
                    programmer = debugger.create_jtag_programmer(jtag)
                    flash_uid = programmer.read_flash_uid()
                    out(f"\tFlash UID: {flash_uid:016x}")
            out("")

        return found_device



#!/usr/bin/env python3
#
# This file is part of Apollo.
#
# Copyright (c) 2020-2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import print_function
from operator import invert

import os
import sys
import ast
import time
import errno
import logging
import argparse
from collections import namedtuple
import xdg.BaseDirectory
from functools import partial
import deprecation

from apollo_fpga import ApolloDebugger, __version__
from apollo_fpga.jtag import JTAGChain, JTAGPatternError
from apollo_fpga.ecp5 import ECP5_JTAGProgrammer, ECP5FlashBridgeProgrammer
from apollo_fpga.onboard_jtag import *


#
# Common JEDEC manufacturer IDs for SPI flash chips.
#
JEDEC_MANUFACTURERS = {
    0x01: "AMD/Spansion/Cypress",
    0x04: "Fujitsu",
    0x1C: "eON",
    0x1F: "Atmel/Microchip",
    0x20: "Micron/Numonyx/ST",
    0x37: "AMIC",
    0x62: "SANYO",
    0x89: "Intel",
    0x8C: "ESMT",
    0xA1: "Fudan",
    0xAD: "Hyundai",
    0xBF: "SST",
    0xC2: "Micronix",
    0xC8: "Gigadevice",
    0xD5: "ISSI",
    0xEF: "Winbond",
    0xE0: 'Paragon',
}

#
# Common JEDEC device IDs. Prefixed with their manufacturer for easy / unique lookup.
#
JEDEC_PARTS = {
    0xEF3015: "W25X16L",
    0xEF3014: "W25X80L",
    0xEF3013: "W25X40L",
    0xEF3012: "W25X20L",
    0xEF3011: "W25X10L",
    0xEF4015: "W25Q16DV",
    0xEF4016: "W25Q32DV",
    0xEF4017: "W25Q64DV",
    0xEF4018: "W25Q128DV",
    0xC22515: "MX25L1635E",
    0xC22017: "MX25L6405D",
    0xC22016: "MX25L3205D",
    0xC22015: "MX25L1605D",
    0xC22014: "MX25L8005",
    0xC22013: "MX25L4005",
    0xC22010: "MX25L512E",
    0x204011: "M45PE10",
    0x202014: "M25P80",
    0x1f4501: "AT24DF081",
    0x1C3114: "EN25F80",
    0xE04014: "PN25F08",
}


@deprecation.deprecated(deprecated_in="1.1.0", removed_in="2.0.0",
                        current_version=__version__,
                        details="Use ApolloDebugger.print_info() instead.")
def print_device_info(device, args):
    """ Command that prints information about devices connected to the scan chain to the console. """

    logging.info(f"Detected a {device.get_compatibility_string()} device!")
    logging.info(f"\tHardware: {device.get_hardware_name()}")
    logging.info(f"\tSerial number: {device.serial_number}")
    logging.info(f"\tFirmware version: {device.get_firmware_version()}")
    logging.info(f"\tUSB API version: {device.get_usb_api_version_string()}")
    if args.force_offline:
        device.force_fpga_offline()
        with device.jtag as jtag:
            programmer = device.create_jtag_programmer(jtag)
            flash_uid = programmer.read_flash_uid()
        logging.info(f"\tFlash UID: {flash_uid:016x}")

def print_chain_info(device, args):
    """ Command that prints information about devices connected to the scan chain to the console. """

    with device.jtag as jtag:
        logging.info("Scanning for connected devices...")
        detected_devices = jtag.enumerate()

        # If devices exist on the scan chain, print their information.
        if detected_devices:
            logging.info("{} device{} detected on the scan chain:\n".format(
                        len(detected_devices), 's' if len(detected_devices) > 1 else ''))

            for device in detected_devices:
                logging.info("    {:08x} -- {}".format(device.idcode(), device.description()))


            logging.info('')

        else:
            logging.info("No devices found.\n")


def play_svf_file(device, args):
    """ Command that prints the relevant flash chip's information to the console. """

    if not args.file:
        logging.error("You must provide an SVF filename to play!\n")
        sys.exit(-1)

    with device.jtag as jtag:
        try:
            jtag.play_svf_file(args.file)
        except JTAGPatternError:
            # Our SVF player has already logged the error to stderr.
            logging.error("")


def configure_fpga(device, args):
    """ Command that prints information about devices connected to the scan chain to the console. """

    with device.jtag as jtag:
        programmer = device.create_jtag_programmer(jtag)

        with open(args.file, "rb") as f:
            bitstream = f.read()

        programmer.configure(bitstream)

    # Let the LUNA gateware take over in devices with shared USB port
    device.allow_fpga_takeover_usb()


def ensure_unconfigured(device):
    with device.jtag as jtag:
        programmer = device.create_jtag_programmer(jtag)
        programmer.unconfigure()


def erase_flash(device, args):
    ensure_unconfigured(device)

    with device.jtag as jtag:
        programmer = device.create_jtag_programmer(jtag)
        programmer.erase_flash()


def program_flash(device, args):
    if args.fast:
        return program_flash_fast(device, args)

    ensure_unconfigured(device)

    with device.jtag as jtag:
        programmer = device.create_jtag_programmer(jtag)
        offset = ast.literal_eval(args.offset) if args.offset else 0

        with open(args.file, "rb") as f:
            bitstream = f.read()

        programmer.flash(bitstream, offset=offset)



def program_flash_fast(device, args):
    # check if flash-fast is supported
    try:
        from amaranth.build.run import LocalBuildProducts
        from luna.gateware.platform import get_appropriate_platform
        from apollo_fpga.gateware.flash_bridge import FlashBridge, FlashBridgeConnection
    except ImportError:
        logging.error("`flash --fast` requires the `luna` package in the Python environment.\n"
                      "Install `luna` or use `flash` instead.")
        sys.exit(-1)

    # get platform
    platform=get_appropriate_platform()

    # Retrieve a FlashBridge cached bitstream or build it
    plan = platform.build(FlashBridge(), do_build=False)
    cache_dir = os.path.join(
        xdg.BaseDirectory.save_cache_path('apollo'), 'build', plan.digest().hex()
    )
    if os.path.exists(cache_dir):
        products = LocalBuildProducts(cache_dir)
    else:
        products = plan.execute_local(cache_dir)

    # Configure flash bridge
    with device.jtag as jtag:
        programmer = device.create_jtag_programmer(jtag)
        programmer.configure(products.get("top.bit"))

    # Let the LUNA gateware take over in devices with shared USB port
    device.allow_fpga_takeover_usb()

    # Program SPI flash memory using the configured bridge
    bridge = FlashBridgeConnection()
    programmer = ECP5FlashBridgeProgrammer(bridge=bridge)
    with open(args.file, "rb") as f:
        bitstream = f.read()
    programmer.flash(bitstream)


def program_flash_fast_deprecated(device, args):
    logging.warning("[WARNING] `flash-fast` is now deprecated. Please use `flash --fast` instead.")
    return program_flash_fast(device, args)


def read_back_flash(device, args):
    ensure_unconfigured(device)

    # XXX abstract this?
    length = ast.literal_eval(args.length) if args.length else (4 * 1024 * 1024)
    offset = ast.literal_eval(args.offset) if args.offset else 0
    if offset:
        length = min(length, 4 * 1024 * 1024 - offset)

    with device.jtag as jtag:
        programmer = device.create_jtag_programmer(jtag)

        with open(args.file, "wb") as f:
            bitstream = programmer.read_flash(length, offset=offset)
            f.write(bitstream)

    device.soft_reset()



def print_flash_info(device, args):
    """ Command that prints information about the currently connected FPGA's configuration flash. """
    device.force_fpga_offline()
    serial_number = device.serial_number

    with device.jtag as jtag:
        programmer = device.create_jtag_programmer(jtag)
        manufacturer, device = programmer.read_flash_id()
        unique_id = programmer.read_flash_uid()

        if manufacturer == 0xFF:
            logging.info("No flash detected.")
            return

        logging.info(f"Device serial number: {serial_number}")
        logging.info("Detected an FPGA-connected SPI configuration flash!")

        try:
            logging.info(f"\tManufacturer: {JEDEC_MANUFACTURERS[manufacturer]} ({manufacturer:02x})")
        except KeyError:
            logging.info(f"\tUnknown manufacturer ({manufacturer:02x}).")

        try:
            logging.info(f"\tDevice: {JEDEC_PARTS[device]} ({device:06x})")
        except KeyError:
            logging.info(f"\tUnknown device ({device:06x}).")

        logging.info(f"\tUnique ID: {unique_id:016x}")

        logging.info("")


def reconfigure_fpga(device, args):
    """ Command that requests the attached ECP5 reconfigure itself from its SPI flash. """
    device.soft_reset()

    # Let the LUNA gateware take over in devices with shared USB port
    device.allow_fpga_takeover_usb()


def force_fpga_offline(device, args):
    """ Command that requests the attached ECP5 be held unconfigured. """
    device.force_fpga_offline()


def _do_debug_spi(device, spi, args, *, invert_cs):

    # Try to figure out what data the user wants to send.
    data_raw = ast.literal_eval(args.bytes)
    if isinstance(data_raw, int):
        data_raw = [data_raw]

    data_to_send = bytes(data_raw)
    response     = spi.transfer(data_to_send, invert_cs=invert_cs)

    print("response: {}".format(response))


def debug_spi(device, args, *, invert_cs=False):
    _do_debug_spi(device, device.spi, args, invert_cs=invert_cs)


def jtag_debug_spi(device, args):
    """ Command that issues data over a JTAG-over-SPI connection. """

    with device.jtag as jtag:
        spi, _ = device.create_jtag_spi(jtag)
        _do_debug_spi(device, spi, args, invert_cs=False)


def set_led_pattern(device, args):
    device.set_led_pattern(int(args.pattern))

def debug_spi_inv(device, args):
    debug_spi(device, args, invert_cs=True)


def _do_debug_spi_register(device, spi, args):

    # Try to figure out what data the user wants to send.
    address = int(args.address, 0)
    if args.value:
        value = int(args.value, 0)
        is_write = True
    else:
        value = 0
        is_write = False

    try:
        response = spi.register_transaction(address, is_write=is_write, value=value)
        print("0x{:08x}".format(response))
    except IOError as e:
        logging.critical(f"{e}\n")


def debug_spi_register(device, args):
    _do_debug_spi_register(device, device.spi, args)

def jtag_debug_spi_register(device, args):
    _reg, reg = device.create_jtag_spi(device.jtag)
    _do_debug_spi_register(device, reg, args)


Command = namedtuple("Command", ("name", "alias", "args", "help", "handler"),
                     defaults=(None, [], [], None, None))

COMMANDS = [
    # Info queries
    Command("info", handler=print_device_info, args=[(("--force-offline",), dict(action='store_true'))],
            help="Print device info.", ),
    Command("jtag-scan", handler=print_chain_info,
            help="Prints information about devices on the onboard JTAG chain."),
    Command("flash-info", handler=print_flash_info,
            help="Prints information about the FPGA's attached configuration flash."),

    # Flash commands
    Command("flash-erase", handler=erase_flash,
            help="Erases the contents of the FPGA's flash memory."),
    Command("flash-program", args=["file", "--offset", (("--fast",), dict(action='store_true'))],
            handler=program_flash, help="Programs the target bitstream onto the FPGA's configuration flash."),
    Command("flash-fast", args=["file", "--offset"], handler=program_flash_fast_deprecated,
            help="Programs a bitstream onto the FPGA's configuration flash using a SPI bridge."),
    Command("flash-read", args=["file", "--offset", "--length"], handler=read_back_flash,
            help="Reads the contents of the attached FPGA's configuration flash."),

    # JTAG commands
    Command("svf", args=["file"], handler=play_svf_file,
            help="Plays a given SVF file over JTAG."),
    Command("configure", args=["file"], handler=configure_fpga,
            help="Uploads a bitstream to the device's FPGA over JTAG."),
    Command("reconfigure", handler=reconfigure_fpga,
            help="Requests the attached ECP5 reconfigure itself from its SPI flash."),
    Command("force-offline", handler=force_fpga_offline,
            help="Forces the board's FPGA offline."),

    # SPI debug exchanges
    Command("spi", args=["bytes"], handler=debug_spi,
            help="Sends the given list of bytes over debug-SPI, and returns the response."),
    Command("spi-inv", args=["bytes"], handler=debug_spi_inv,
            help="Sends the given list of bytes over SPI with inverted CS."),
    Command("spi-reg", args=["address", "value"], handler=debug_spi_register,
            help="Reads or writes to a provided register over the debug-SPI."),

    # JTAG-SPI debug exchanges.
    Command("jtag-spi", args=["bytes"], handler=jtag_debug_spi,
            help="Sends the given list of bytes over SPI-over-JTAG, and returns the response."),
    Command("jtag-reg", args=["address", "value"], handler=jtag_debug_spi_register,
            help="Reads or writes to a provided register of JTAG-tunneled debug SPI."),

    # Misc
    Command("leds", args=["pattern"], handler=set_led_pattern,
            help="Sets the specified pattern for the Debug LEDs."),
]


def main():

    # Set up a simple argument parser.
    parser = argparse.ArgumentParser(description="Apollo FPGA Configuration / Debug tool",
            formatter_class=argparse.RawTextHelpFormatter)
    sub_parsers = parser.add_subparsers(dest="command", metavar="command")
    for command in COMMANDS:
        cmd_parser = sub_parsers.add_parser(command.name, aliases=command.alias, help=command.help)
        cmd_parser.set_defaults(func=command.handler)
        for arg in command.args:
            if isinstance(arg, tuple):
                cmd_parser.add_argument(*arg[0], **arg[1])
            else:
                cmd_parser.add_argument(arg)

    args = parser.parse_args()
    if not args.command:
        parser.print_help()
        return

    # Force the FPGA offline by default in most commands to force Apollo mode if needed.
    force_offline = args.force_offline if "force_offline" in args else True

    # Set up python's logging to act as a simple print, for now.
    logging.basicConfig(level=logging.INFO, format="%(message)-s")

    if args.command == "info":
        if ApolloDebugger.print_info(force_offline=force_offline, out=logging.info):
            if not force_offline:
                logging.info(f"For additional device information use the --force-offline option.")
        return

    device = ApolloDebugger(force_offline=force_offline)

    # Execute the relevant command.
    args.func(device, args)


if __name__ == '__main__':
    main()



#
# This file is part of Apollo.
#
# Copyright (c) 2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

import usb.core

from amaranth                          import Signal, Elaboratable, Module, Cat, C
from amaranth.lib.fifo                 import AsyncFIFO

from luna.gateware.interface.flash     import ECP5ConfigurationFlashInterface
from luna.gateware.interface.spi       import SPIBus
from luna.gateware.stream              import StreamInterface
from luna.gateware.usb.usb2.request    import USBRequestHandler
from luna.gateware.usb.request.windows import MicrosoftOS10DescriptorCollection
from luna.gateware.usb.request.windows import MicrosoftOS10RequestHandler
from luna.usb2                         import USBDevice, USBStreamInEndpoint, USBStreamOutEndpoint

from usb_protocol.types                import USBRequestType, USBRequestRecipient
from usb_protocol.emitters             import DeviceDescriptorCollection
from usb_protocol.emitters.descriptors.standard import get_string_descriptor

from apollo_fpga                       import ApolloDebugger
from .advertiser                       import ApolloAdvertiser, ApolloAdvertiserRequestHandler

VENDOR_ID  = 0x1209
PRODUCT_ID = 0x000F

BULK_ENDPOINT_NUMBER = 1
MAX_BULK_PACKET_SIZE = 512


class SPIStreamController(Elaboratable):
    """ Class that drives a SPI bus with data from input stream packets.
    Data received from the device is returned as another packet."""

    def __init__(self):
        self.period = 4  # powers of two only
        self.bus    = SPIBus()
        self.input  = StreamInterface()
        self.output = StreamInterface()

    def elaborate(self, platform):
        m = Module()

        # Counter for clock generation
        cycles = Signal(range(self.period))

        # Generate strobes for clock edges
        sck_fall = Signal()
        sck_rise = Signal()
        sck_d    = Signal()
        m.d.sync += sck_d.eq(self.bus.sck)
        m.d.comb += [
            sck_fall.eq( sck_d & ~self.bus.sck),  # falling edge
            sck_rise.eq(~sck_d &  self.bus.sck),  # rising edge
        ]

        # I/O shift registers, bit counter and last flag
        shreg_o = Signal(8)
        shreg_i = Signal(8)
        count_o = Signal(range(8))
        last    = Signal()

        m.d.comb += [
            self.bus.sdi        .eq(shreg_o[-1]),
            self.output.payload .eq(shreg_i),
        ]

        with m.FSM() as fsm:
            m.d.comb += self.bus.cs.eq(~fsm.ongoing('IDLE'))

            with m.State("IDLE"):
                m.d.comb += [
                    self.input.ready    .eq(1),
                    self.bus.sck        .eq(0),
                ]
                with m.If(self.input.valid):
                    m.next = 'SHIFT'

            with m.State("WAIT"):
                m.d.comb += [
                    self.input.ready    .eq(1),
                    self.bus.sck        .eq(0),
                ]
                with m.If(self.input.valid):
                    m.next = 'SHIFT'

            with m.State("SHIFT"):
                m.d.comb += [
                    self.input.ready    .eq(sck_fall & (count_o == 0) & ~last),
                    self.bus.sck        .eq(cycles[-1])
                ]
                m.d.sync += cycles.eq(cycles + 1)

                # Read logic, latch on rising edge
                m.d.sync += self.output.valid.eq(0)
                with m.If(sck_rise):
                    m.d.sync += [
                        shreg_i             .eq(Cat(self.bus.sdo, shreg_i[:-1])),
                        self.output.valid   .eq(count_o == 0),
                        self.output.last    .eq(last),
                    ]

                # Write logic, setup on falling edge
                with m.If(sck_fall):
                    m.d.sync += [
                        shreg_o             .eq(Cat(C(0,1), shreg_o[:-1])),
                        count_o             .eq(count_o - 1),
                    ]
                    with m.If(count_o == 0):
                        with m.If(last):
                            m.next = 'END'
                        with m.Elif(~self.input.valid):
                            m.next = 'WAIT'

            with m.State("END"):
                m.d.comb += [
                    self.input.ready    .eq(0),
                    self.bus.sck        .eq(0),
                ]
                m.d.sync += [
                    last    .eq(0),
                    cycles  .eq(0),
                ]
                m.next = 'IDLE'

        with m.If(self.input.valid & self.input.ready):
            m.d.sync += [
                shreg_o     .eq(self.input.payload),
                last        .eq(self.input.last),
                count_o     .eq(7),
            ]

        return m


class FlashBridgeRequestHandler(USBRequestHandler):
    """ Request handler that can trigger a FPGA reconfiguration. """

    REQUEST_TRIGGER_RECONF = 0

    def __init__(self, if_number):
        super().__init__()
        self.if_number = if_number

    def elaborate(self, platform):
        m = Module()

        interface         = self.interface
        setup             = self.interface.setup

        #
        # Vendor request handlers.

        self_prog = platform.request("self_program", dir="o").o

        with m.If((setup.type == USBRequestType.VENDOR) & \
                  (setup.recipient == USBRequestRecipient.INTERFACE) & \
                  (setup.index == self.if_number)):

            with m.Switch(setup.request):

                with m.Case(self.REQUEST_TRIGGER_RECONF):

                    m.d.comb += interface.claim.eq(1)

                    # Once the receive is complete, respond with an ACK.
                    with m.If(interface.rx_ready_for_response):
                        m.d.comb += interface.handshakes_out.ack.eq(1)

                    # If we reach the status stage, send a ZLP.
                    with m.If(interface.status_requested):
                        m.d.comb += self.send_zlp()
                        m.d.usb += self_prog.eq(1)

        return m


class FlashBridgeSubmodule(Elaboratable):
    """ Implements gateware for the USB<->SPI bridge. Intended to use as a submodule
        See example in FlashBridge """

    def __init__(self, endpoint):
        # Endpoint number for the in/out stream endpoints
        self.endpoint = endpoint

        # Define endpoints
        self.endpoint_out = USBStreamOutEndpoint(
            endpoint_number=endpoint,
            max_packet_size=MAX_BULK_PACKET_SIZE,
        )
        self.endpoint_in = USBStreamInEndpoint(
            endpoint_number=endpoint,
            max_packet_size=MAX_BULK_PACKET_SIZE
        )

    def elaborate(self, platform):
        m = Module()

        stream_in  = self.endpoint_in.stream
        stream_out = self.endpoint_out.stream

        # Use two small asynchronous FIFOs for crossing clock domains
        spi     = SPIStreamController()
        spi_bus = ECP5ConfigurationFlashInterface(bus=platform.request('spi_flash'), use_cs=True)
        tx_fifo = AsyncFIFO(width=8+1, depth=8, w_domain="usb", r_domain="sync")
        rx_fifo = AsyncFIFO(width=8+1, depth=8, w_domain="sync", r_domain="usb")

        m.submodules += spi
        m.submodules += spi_bus
        m.submodules += tx_fifo
        m.submodules += rx_fifo

        m.d.comb += [
            # Connect output from USB host to transmission FIFO
            tx_fifo.w_data      .eq(Cat(stream_out.payload, stream_out.last)),
            tx_fifo.w_en        .eq(stream_out.valid),
            stream_out.ready    .eq(tx_fifo.w_rdy),

            # Connect transmission FIFO to the SPI controller
            Cat(spi.input.payload, spi.input.last).eq(tx_fifo.r_data),
            spi.input.valid     .eq(tx_fifo.r_rdy),
            tx_fifo.r_en        .eq(spi.input.ready),

            # Connect output from SPI controller to reception FIFO
            rx_fifo.w_data      .eq(Cat(spi.output.payload, spi.output.last)),
            rx_fifo.w_en        .eq(spi.output.valid),
            spi.output.ready    .eq(1),  # ignore rx_fifo.w_rdy

            # Connect reception FIFO to USB host input
            Cat(stream_in.payload, stream_in.last).eq(rx_fifo.r_data),
            stream_in.valid     .eq(rx_fifo.r_rdy),
            rx_fifo.r_en        .eq(stream_in.ready),

            # Connect the SPI bus to our SPI controller
            spi_bus.sck         .eq(spi.bus.sck),
            spi_bus.sdi         .eq(spi.bus.sdi),
            spi_bus.cs          .eq(spi.bus.cs),
            spi.bus.sdo         .eq(spi_bus.sdo),
        ]

        return m


class FlashBridge(Elaboratable):

    def create_descriptors(self, sharing):
        """ Create the descriptors we want to use for our device. """

        descriptors = DeviceDescriptorCollection()

        #
        # We'll add the major components of the descriptors we we want.
        # The collection we build here will be necessary to create a standard endpoint.
        #

        # We'll need a device descriptor...
        with descriptors.DeviceDescriptor() as d:
            d.idVendor           = VENDOR_ID
            d.idProduct          = PRODUCT_ID

            d.iManufacturer      = "Apollo Project"
            d.iProduct           = "Configuration Flash Bridge"

            d.bNumConfigurations = 1

        # ... and a description of the USB configuration we'll provide.
        with descriptors.ConfigurationDescriptor() as c:

            with c.InterfaceDescriptor() as i:
                i.bInterfaceNumber = 0
                i.bInterfaceClass = 0xFF
                i.bInterfaceSubclass = 0x01
                i.bInterfaceProtocol = 0x00

                with i.EndpointDescriptor() as e:
                    e.bEndpointAddress = BULK_ENDPOINT_NUMBER
                    e.wMaxPacketSize   = MAX_BULK_PACKET_SIZE

                with i.EndpointDescriptor() as e:
                    e.bEndpointAddress = 0x80 | BULK_ENDPOINT_NUMBER
                    e.wMaxPacketSize   = MAX_BULK_PACKET_SIZE

            # If sharing the port, add the Apollo stub interface.
            if sharing is not None:
                with c.InterfaceDescriptor() as i:
                    i.bInterfaceNumber = 1
                    i.bInterfaceClass = 0xFF
                    i.bInterfaceSubclass = 0x00
                    i.bInterfaceProtocol = ApolloAdvertiserRequestHandler.PROTOCOL_VERSION

        return descriptors

    def elaborate(self, platform):
        m = Module()

        # Generate our domain clocks/resets.
        m.submodules.car = platform.clock_domain_generator()

        # Create our USB device interface...
        phy_name = platform.apollo_gateware_phy
        ulpi = platform.request(phy_name)
        m.submodules.usb = usb = USBDevice(bus=ulpi)

        # Check how the port is shared with Apollo.
        sharing = platform.port_sharing(phy_name)

        # Create descriptors.
        descriptors = self.create_descriptors(sharing)

        # Add Microsoft OS 1.0 descriptors for Windows compatibility.
        descriptors.add_descriptor(get_string_descriptor("MSFT100\xee"), index=0xee)
        msft_descriptors = MicrosoftOS10DescriptorCollection()
        with msft_descriptors.ExtendedCompatIDDescriptor() as c:
            with c.Function() as f:
                f.bFirstInterfaceNumber = 0
                f.compatibleID          = 'WINUSB'
            if sharing is not None:
                with c.Function() as f:
                    f.bFirstInterfaceNumber = 1
                    f.compatibleID          = 'WINUSB'

        # Add our standard control endpoint to the device.
        control_ep = usb.add_standard_control_endpoint(descriptors, avoid_blockram=True)

        # Add handler for Microsoft descriptors.
        msft_handler = MicrosoftOS10RequestHandler(msft_descriptors, request_code=0xee)
        control_ep.add_request_handler(msft_handler)

        # Add our vendor request handler to the control endpoint.
        control_ep.add_request_handler(FlashBridgeRequestHandler(0))

        # If needed, create an advertiser and add its request handler.
        if sharing == "advertising":
            adv = m.submodules.adv = ApolloAdvertiser()
            control_ep.add_request_handler(adv.default_request_handler(1))

        # Add bridge submodule and input/output stream endpoints to our device.
        m.submodules.bridge = bridge = FlashBridgeSubmodule(BULK_ENDPOINT_NUMBER)
        usb.add_endpoint(bridge.endpoint_in)
        usb.add_endpoint(bridge.endpoint_out)

        # Connect our device
        m.d.comb += usb.connect.eq(1)

        return m


class FlashBridgeNotFound(IOError):
    pass

class FlashBridgeConnection:
    def __init__(self):
        # Try to create a connection to our configuration flash bridge.
        device = ApolloDebugger._find_device(
            ids=[(VENDOR_ID, PRODUCT_ID)],
            custom_match=self._find_cfg_flash_bridge,
            timeout=5000
        )

        # If we couldn't find the bridge, bail out.
        if device is None:
            raise FlashBridgeNotFound("Unable to find device")

        self.device = device
        self.interface, self.endpoint = self._find_cfg_flash_bridge(device, get_ep=True)

    def __del__(self):
        self.request_handoff()

    @staticmethod
    def _find_cfg_flash_bridge(dev, get_ep=False):
        for cfg in dev:
            for intf in usb.util.find_descriptor(cfg, find_all=True, bInterfaceClass=0xFF, bInterfaceSubClass=0x01):
                if not get_ep:
                    return True
                return intf.bInterfaceNumber, intf[0].bEndpointAddress
        return None, None if get_ep else False

    def trigger_reconfiguration(self):
        """ Triggers the target FPGA to reconfigure itself from its flash chip. """
        request_type = usb.ENDPOINT_OUT | usb.RECIP_INTERFACE | usb.TYPE_VENDOR
        return self.device.ctrl_transfer(request_type, 0, wValue=0, wIndex=self.interface)

    def request_handoff(self):
        """ Requests the gateware to liberate the USB port. """
        try:
            ApolloDebugger._request_handoff(self.device)
        except Exception:
            pass

    def transfer(self, data):
        """ Performs a SPI transfer, targeting the configuration flash."""
        tx_sent = self.device.write(self.endpoint, data)
        assert tx_sent == len(data)
        rx_data = self.device.read(0x80 | self.endpoint, 512)
        assert len(rx_data) == tx_sent, f'Expected {tx_sent} bytes, received {len(rx_data)}'
        return rx_data
